#include "balancer"
#include "../tcpdispatcher/tcpdispatcher"
#include "../httpdispatcher/httpdispatcher"

void Balancer::serve() {
    int clsock;

    // Start up wakeup/checkup handlers.
    if (config.wakeupsec() && !config.foregroundmode() && config.sport()) {
	msg ("Starting wakeup thread.\n");
	Wakeupthread *wt = new Wakeupthread();
	wt->start();
    }
    if (config.checkupsec() && !config.foregroundmode() && config.sport()) {
	msg ("Starting checkup thread.\n");
	Checkupthread *ct = new Checkupthread();
	ct->start();
    }

    // Write the PID file.
    if (config.pidfile() != "") {
	FILE *f;
	if (! (f = fopen (config.pidfile().c_str(), "w")) )
	    throw ((Error) "Cannot write pid file " + config.pidfile() +
		   ": " + strerror(errno));
	fprintf (f, "%u\n", getpid());
	fclose (f);
    }

    // Wait for activity, serve it.
    ostringstream o;
    o << server_fd;
    msg ("Awaiting activity on fd " + o.str() +"\n");
    while (1) {
	Fdset fdset(0);
	fdset.add (server_fd);
	if (fdset.readable() < 0) {
	    // We caught a signal. Either a request to report status,
	    // or to terminate.
	    msg ("Interrupt seen\n");
	    if (terminate()) {
		msg ("Termination requested, XR will stop.\n");
		close (server_fd);
		shutdown (server_fd, SHUT_RDWR);
		unsigned prev_conn = 0x19081962;
		while (1) {
		    unsigned curr_conn = balancer.connections();
		    if (!curr_conn)
			break;
		    if (curr_conn != prev_conn) {
			ostringstream o;
			o << curr_conn;
			msg ("There are still " + o.str() + " connections\n");
			prev_conn = curr_conn;
		    }
		    sleep (1);
		}
		msg ("XR is idle, stopping.\n");
		break;
	    } else if (report()) {
		msg ("Report requested\n");
		reportmsg ("*** XR STATUS REPORT STARTS ***\n");
		for (unsigned i = 0; i < nbackends(); i++) {
		    reportmsg ("Back end " + backend(i).description() + ":\n");
		    reportmsg ("  Status: " + backend(i).availablestr() +
			       ", " + backend(i).livestr() + "\n");
		    ostringstream o;
		    o << backend(i).connections() << " (max "
		      << backend(i).maxconn() << ")";
		    reportmsg ("  Connections: " + o.str() + "\n");
		    ostringstream b;
		    b << backend(i).bytesserved() << " bytes, "
		      << backend(i).clientsserved() << " clients";
		    reportmsg ("  Served: " + b.str() + "\n");
		}
		report (false);
		reportmsg ("*** XR STATUS REPORT ENDS ***\n");
		continue;
	    } else {
		msg ("Non-meaningful interrupt, resuming\n");
		continue;
	    }
	}
	
	if (server_fd) {
	    // If tcp-serving: server_fd > 0; serve and loop again
	    int size;
	    struct sockaddr_in clname;

	    size = sizeof(clname);
	    if ( (clsock = accept (server_fd, (struct sockaddr *) &clname,
				   (socklen_t*) &size)) < 0 )
		throw ((Error)"Failed to accept network connection");
	    
	    string clientip = inet_ntoa(clname.sin_addr);

	    // If there is an allow list, the client must match it.
	    if (config.nallow()) {
		debugmsg ("Matching " + clientip + " against allow list\n");
		bool allowed = false;
		for (unsigned n = 0; n < config.nallow(); n++) {
		    if (ipmatch (clname.sin_addr, config.allow(n))) {
			allowed = true;
			break;
		    }
		}
		if (!allowed) {
		    msg ("Not serving client " + clientip +
			 ": no match in allow list\n");
		    close (clsock);
		    continue;
		}
	    }
	    // If the client is in the deny list, deny it.
	    debugmsg ("Matching " + clientip + " against deny list\n");
	    bool denied = false;
	    for (unsigned n = 0; n < config.ndeny(); n++)
		if (ipmatch (clname.sin_addr, config.deny(n))) {
		    denied = true;
		    break;
		}
	    if (denied) {
		msg ("Not serving client " + clientip +
		     ": match in deny list\n");
		close (clsock);
		continue;
	    }

	    // Show whom we've accepted
	    if (config.verbose()) {
		ostringstream o;
		o << clsock;
		msg ("Accepted connection from " + clientip +
		     " as client fd " + o.str() +"\n");
		msg ("Current back end states:\n");
		for (unsigned i = 0; i < nbackends(); i++) {
		    ostringstream status;
		    status << backend(i).connections() << " connections";
		    if (backend(i).maxconn())
			status << " (max " << backend(i).maxconn() << ")";
		    status << ", status " <<  backend(i).availablestr();
		    msg ("Back end " + backend(i).description() + ": " +
			 status.str() + "\n");
		    ostringstream n;
		    n << connections();
		    msg ("Balancer is serving " + n.str() + " clients\n");
		}
	    }
	    
	    // We got action! Check if the total connections to the
	    // balancer doesn't exceed the max.
	    if (config.maxconn() && connections() >= config.maxconn()) {
		ostringstream o;
		o << connections() << " connections (max "
		  << config.maxconn() << ")";
		msg ("Not serving client: already " + o.str() + "\n");
		close (clsock);
		continue;
	    }

	    TcpDispatcher *d;
	    switch (config.stype()) {
	    case Servertype::t_tcp:
		d = new TcpDispatcher (clsock, clientip);
		break;
	    case Servertype::t_http:
		d = new HttpDispatcher (clsock, clientip);
		break;
	    default:
		throw ((Error) "Internal error, can't choose dispatcher");
		break;
	    }

	    // Allocation boundary printout
	    if (config.debug()) {
		void *mem = malloc(16);
		free (mem);
		ostringstream o;
		o << mem;
		debugmsg ("Allocation boundary at dispatcher start: " +
			  o.str() + "\n");
	    }
	    
	    d->start();
	} else {
	    // If fd-serving, serve and close. Don't thread it up.
	    TcpDispatcher *d;
	    switch (config.stype()) {
	    case Servertype::t_tcp:
		d = new TcpDispatcher (server_fd, "0.0.0.0");
		break;
	    case Servertype::t_http:
		d = new HttpDispatcher (server_fd, "0.0.0.0");
		break;
	    default:
		throw ((Error) "Internal error, can't choose dispatcher");
		break;
	    }
	    d->execute();
	    break;
	}
    }

    // If a PID stamp was created, remove it now.
    if (config.pidfile() != "")
	unlink (config.pidfile().c_str());
}
